@@ -5,9 +5,9 @@ import numpy as np
 import SimpleITK as sitk
 import cv2
 import geojson
+from wsireg.reg_transform import RegTransform
 from wsireg.utils.tform_utils import (
     wsireg_transforms_to_itk_composite,
-    collate_wsireg_transforms,
 )
 
 GJ_SHAPE_TYPE = {
@@ -43,8 +43,16 @@ def gj_to_np(gj: dict):
             pts.append(np.squeeze(np.array(geo)))
         pts = np.vstack(pts)
 
-    else:
+    elif gj.get("geometry").get("type") == "Polygon":
         pts = np.squeeze(np.asarray(gj.get("geometry").get("coordinates")))
+    elif gj.get("geometry").get("type") == "Point":
+        pts = np.expand_dims(
+            np.asarray(gj.get("geometry").get("coordinates")), 0
+        )
+    elif gj.get("geometry").get("type") == "MultiPoint":
+        pts = np.asarray(gj.get("geometry").get("coordinates"))
+    elif gj.get("geometry").get("type") == "LineString":
+        pts = np.asarray(gj.get("geometry").get("coordinates"))
 
     if gj.get("properties").get("classification") is None:
         shape_name = "unnamed"
@@ -215,9 +223,7 @@ def scale_shape_coordinates(poly: dict, scale_factor: float):
     return poly
 
 
-def invert_nonlinear_transforms(
-    itk_transforms: list, elx_transform_data: list
-):
+def invert_nonlinear_transforms(itk_transforms: list):
     """
     Check list of sequential ITK transforms for non-linear (i.e., bspline) transforms
     Transformations need to be inverted to transform from moving to fixed space as transformations
@@ -230,44 +236,31 @@ def invert_nonlinear_transforms(
     ----------
     itk_transforms:list
         list of itk.Transform
-    elx_transform_data: list
-        list of dict containing elastix transformation parameters
+
     Returns
     -------
     itk_transforms:list
         list of itk.Transform where any non-linear transforms are replaced with an inverted displacement field
     """
-    tform_linear = [t.IsLinear() for t in itk_transforms]
+    tform_linear = [t.is_linear for t in itk_transforms]
 
     if all(tform_linear):
         return itk_transforms
     else:
         nl_idxs = np.where(np.array(tform_linear) == 0)[0]
-        print(
-            f"transform(s) at index {[n for n in nl_idxs]} is(are) non-linear, inverting displacement field(s)\n"
-            "this can take some time"
-        )
         for nl_idx in nl_idxs:
-            tform = elx_transform_data[4]
-            tform_to_dfield = sitk.TransformToDisplacementFieldFilter()
-            tform_to_dfield.SetOutputSpacing(
-                [float(p) for p in tform["Spacing"]]
-            )
-            tform_to_dfield.SetOutputOrigin(
-                [float(p) for p in tform["Origin"]]
-            )
-            tform_to_dfield.SetOutputDirection(
-                [float(p) for p in tform["Direction"]]
-            )
-            tform_to_dfield.SetSize([int(p) for p in tform["Size"]])
-            dfield = tform_to_dfield.Execute(itk_transforms[nl_idx])
-            dfield = sitk.InvertDisplacementField(dfield)
-            dfield = sitk.DisplacementFieldTransform(dfield)
-            itk_transforms[nl_idx] = dfield
+            if itk_transforms[nl_idx].inverse_transform is None:
+                print(
+                    f"transform at index {nl_idx} is non-linear and the inverse has not been computed\n"
+                    "inverting displacement field(s)...\n"
+                    "this can take some time"
+                )
+                itk_transforms[nl_idx].compute_inverse_nonlinear()
+
     return itk_transforms
 
 
-def prepare_pt_transformation_data(transformations):
+def prepare_pt_transformation_data(transformations, compute_inverse=True):
     """
     Read and prepare wsireg transformation data for point set transformation
 
@@ -276,6 +269,8 @@ def prepare_pt_transformation_data(transformations):
     transformations
         list of dict containing elastix transformation data or str to wsireg .json file containing
         elastix transformation data
+    compute_inverse : bool
+        whether or not to compute the inverse transformation for moving to fixed point transformations
     Returns
     -------
     itk_pt_transforms:list
@@ -284,15 +279,14 @@ def prepare_pt_transformation_data(transformations):
         physical spacing of the final transformation in the transform sequence
         This is needed to map coordinates defined as pixel indices to physical coordinates and then back
     """
-    composite, itk_transforms = wsireg_transforms_to_itk_composite(
-        transformations
-    )
-    elx_transform_data = collate_wsireg_transforms(transformations)
-    itk_pt_transforms = invert_nonlinear_transforms(
-        itk_transforms, elx_transform_data
-    )
-    target_res = float(elx_transform_data[-1]["Spacing"][0])
-    return itk_pt_transforms, target_res
+    if all([isinstance(t, RegTransform) for t in transformations]) is False:
+        composite, transformations = wsireg_transforms_to_itk_composite(
+            transformations
+        )
+    if compute_inverse:
+        transformations = invert_nonlinear_transforms(transformations)
+    target_res = float(transformations[-1].output_spacing[0])
+    return transformations, target_res
 
 
 def itk_transform_pts(
@@ -340,17 +334,9 @@ def itk_transform_pts(
             pt = pt * source_res
         for idx, t in enumerate(itk_transforms):
             if idx == 0:
-                t_pt = (
-                    t.GetInverse().TransformPoint(pt)
-                    if t.IsLinear()
-                    else t.TransformPoint(pt)
-                )
+                t_pt = t.inverse_transform.TransformPoint(pt)
             else:
-                t_pt = (
-                    t.GetInverse().TransformPoint(t_pt)
-                    if t.IsLinear()
-                    else t.TransformPoint(t_pt)
-                )
+                t_pt = t.inverse_transform.TransformPoint(t_pt)
         t_pt = np.array(t_pt)
         if output_idx is True:
             t_pt *= 1 / target_res
@@ -426,9 +412,22 @@ def insert_transformed_pts_gj(gj_data: list, np_data: list):
     shape_gj : dict
         dict of GeoJSON information with updated coordinate information
     """
+
     gj_data_t = deepcopy(gj_data)
     for sh, gj in zip(np_data, gj_data_t):
-        gj.get("geometry").update({"coordinates": [sh["array"].tolist()]})
+        shape_type = gj.get("geometry").get("type")
+
+        if shape_type == "Polygon":
+            gj.get("geometry").update({"coordinates": [sh["array"].tolist()]})
+        elif shape_type == "Point":
+            gj.get("geometry").update(
+                {"coordinates": np.squeeze(sh["array"]).tolist()}
+            )
+        elif shape_type == "MultiPoint":
+            gj.get("geometry").update({"coordinates": sh["array"].tolist()})
+        elif shape_type == "LineString":
+            gj.get("geometry").update({"coordinates": sh["array"].tolist()})
+
     return gj_data_t
 
 
