@@ -1,9 +1,13 @@
 import time
 import yaml
 from pathlib import Path
+from copy import deepcopy
 import json
 import SimpleITK as sitk
-from wsireg.reg_image import RegImage, TransformRegImage
+from wsireg.reg_images.loader import reg_image_loader
+from wsireg.reg_images import MergeRegImage
+from wsireg.writers.merge_ome_tiff_writer import OmeTiffWriter
+from wsireg.writers.merge_ome_tiff_writer import MergeOmeTiffWriter
 from wsireg.utils.reg_utils import (
     register_2d_images_itkelx,
     sitk_pmap_to_dict,
@@ -11,7 +15,15 @@ from wsireg.utils.reg_utils import (
     json_to_pmap_dict,
 )
 from wsireg.reg_shapes import RegShapes
+from wsireg.reg_transform import RegTransform
 from wsireg.utils.config_utils import parse_check_reg_config
+from wsireg.utils.tform_conversion import get_elastix_transforms
+from wsireg.utils.tform_utils import (
+    prepare_wsireg_transform_data,
+)
+from wsireg.utils.im_utils import (
+    ARRAYLIKE_CLASSES,
+)
 
 
 class WsiReg2D(object):
@@ -68,7 +80,6 @@ class WsiReg2D(object):
         project_name: str,
         output_dir: str,
         cache_images=True,
-        reload_data=False,
         config=None,
     ):
 
@@ -83,6 +94,8 @@ class WsiReg2D(object):
         self.image_cache = self.output_dir / ".imcache_{}".format(project_name)
         self.cache_images = cache_images
 
+        self.pairwise = False
+
         self._modalities = {}
         self._modality_names = []
         self._reg_paths = {}
@@ -99,6 +112,9 @@ class WsiReg2D(object):
         self._shape_sets = {}
         self._shape_set_names = []
 
+        self.merge_modalities = {}
+        self.original_size_transforms = {}
+
         if config is not None:
             self.add_data_from_config(config)
 
@@ -367,8 +383,12 @@ class WsiReg2D(object):
 
         for k, v in reg_paths.items():
             tform_path = self.find_path(k, v[-1])
+            if self.pairwise is True:
+                tform_path_modalities = tform_path[:1]
+            else:
+                tform_path_modalities = tform_path
             transform_edges = []
-            for modality in tform_path:
+            for modality in tform_path_modalities:
                 for edge in self.reg_graph_edges:
                     edge_modality = edge["modalities"]['source']
                     if modality == edge_modality:
@@ -400,6 +420,10 @@ class WsiReg2D(object):
         cache_transform_fp = cache_im_fp.parent / "{}_init_tforms.json".format(
             cache_im_fp.stem
         )
+        cache_osize_tform_fp = (
+            self.image_cache
+            / "{}_orig_size_tform.json".format(cache_im_fp.stem)
+        )
 
         if cache_im_fp.exists() is True:
             im_fp = str(cache_im_fp)
@@ -413,7 +437,12 @@ class WsiReg2D(object):
         else:
             im_initial_transforms = None
 
-        return im_fp, im_initial_transforms, im_from_cache
+        if cache_osize_tform_fp.exists() is True:
+            osize_tform = json_to_pmap_dict(cache_osize_tform_fp)
+        else:
+            osize_tform = None
+
+        return im_fp, im_initial_transforms, im_from_cache, osize_tform
 
     def _prepare_modality(self, modality_name, reg_edge, src_or_tgt):
         mod_data = self.modalities[modality_name].copy()
@@ -434,6 +463,7 @@ class WsiReg2D(object):
                 mod_data["preprocessing"],
                 None,
                 mod_data["mask"],
+                None,
             )
         else:
 
@@ -441,9 +471,12 @@ class WsiReg2D(object):
                 mod_data["image_filepath"],
                 mod_data["transforms"],
                 im_from_cache,
+                original_size_transform,
             ) = self._check_cache_modality(modality_name)
 
             if im_from_cache is True:
+                if mod_data["preprocessing"].get("use_mask") is False:
+                    mod_data["mask"] = None
                 mod_data["preprocessing"] = None
 
             return (
@@ -452,6 +485,7 @@ class WsiReg2D(object):
                 mod_data["preprocessing"],
                 mod_data["transforms"],
                 mod_data["mask"],
+                original_size_transform,
             )
 
     def _cache_images(self, modality_name, reg_image):
@@ -462,8 +496,12 @@ class WsiReg2D(object):
             cache_im_fp.stem
         )
 
+        cache_osize_tform_fp = (
+            self.image_cache
+            / "{}_orig_size_tform.json".format(cache_im_fp.stem)
+        )
         if cache_im_fp.is_file() is False:
-            sitk.WriteImage(reg_image.image, str(cache_im_fp), True)
+            sitk.WriteImage(reg_image.reg_image, str(cache_im_fp), True)
 
         if reg_image.mask is not None:
             cache_mask_im_fp = self.image_cache / "{}_prepro_mask.tiff".format(
@@ -473,14 +511,34 @@ class WsiReg2D(object):
                 sitk.WriteImage(reg_image.mask, str(cache_mask_im_fp), True)
 
         if cache_transform_fp.is_file() is False:
-            pmap_dict_to_json(reg_image.transforms, str(cache_transform_fp))
+            pmap_dict_to_json(
+                reg_image.pre_reg_transforms, str(cache_transform_fp)
+            )
+
+        if (
+            cache_osize_tform_fp.is_file() is False
+            and reg_image.original_size_transform is not None
+        ):
+            pmap_dict_to_json(
+                reg_image.original_size_transform, str(cache_osize_tform_fp)
+            )
 
     def _find_nonreg_modalities(self):
         registered_modalities = [
             edge.get("modalities").get("source")
             for edge in self.reg_graph_edges
         ]
-        return list(set(self.modality_names).difference(registered_modalities))
+        non_reg_modalities = list(
+            set(self.modality_names).difference(registered_modalities)
+        )
+
+        # remove attachment modalities
+        for attachment_modality in self.attachment_images.keys():
+            non_reg_modalities.pop(
+                non_reg_modalities.index(attachment_modality)
+            )
+
+        return non_reg_modalities
 
     def save_config(self, registered=False):
         ts = time.strftime('%Y%m%d-%H%M%S')
@@ -507,15 +565,33 @@ class WsiReg2D(object):
                 }
             )
 
+        reg_graph_edges = deepcopy(self.reg_graph_edges)
+
+        [rge.pop("reg_transforms", None) for rge in reg_graph_edges]
+
+        modalities_out = deepcopy(self.modalities)
+        for mod, data in modalities_out.items():
+            if isinstance(data["image_filepath"], ARRAYLIKE_CLASSES):
+                data["image_filepath"] = "ArrayLike"
+
         config = {
             "project_name": self.project_name,
             "output_dir": str(self.output_dir),
             "cache_images": self.cache_images,
-            "modalities": self.modalities,
+            "modalities": modalities_out,
             "reg_paths": reg_paths,
-            "reg_graph_edges": self.reg_graph_edges
+            "reg_graph_edges": reg_graph_edges
             if status == "registered"
             else None,
+            "original_size_transforms": self.original_size_transforms
+            if status == "registered"
+            else None,
+            "attachment_shapes": self.shape_sets
+            if len(self._shape_sets) > 0
+            else None,
+            "attachment_images": self.attachment_images
+            if len(self.attachment_images) > 0
+            else None,
         }
 
         output_path = (
@@ -554,6 +630,7 @@ class WsiReg2D(object):
                     src_prepro,
                     src_transforms,
                     src_mask,
+                    src_original_size_transform,
                 ) = self._prepare_modality(src_name, reg_edge, "source")
 
                 (
@@ -562,24 +639,36 @@ class WsiReg2D(object):
                     tgt_prepro,
                     tgt_transforms,
                     tgt_mask,
+                    tgt_original_size_transform,
                 ) = self._prepare_modality(tgt_name, reg_edge, "target")
 
-                src_reg_image = RegImage(
+                src_reg_image = reg_image_loader(
                     src_reg_image_fp,
                     src_res,
-                    src_prepro,
-                    src_transforms,
+                    preprocessing=src_prepro,
+                    pre_reg_transforms=src_transforms,
                     mask=src_mask,
                 )
 
-                tgt_reg_image = RegImage(
+                tgt_reg_image = reg_image_loader(
                     tgt_reg_image_fp,
                     tgt_res,
-                    tgt_prepro,
-                    tgt_transforms,
+                    preprocessing=tgt_prepro,
+                    pre_reg_transforms=tgt_transforms,
                     mask=tgt_mask,
                 )
 
+                src_reg_image.read_reg_image()
+                tgt_reg_image.read_reg_image()
+
+                if (
+                    tgt_original_size_transform is None
+                    and tgt_reg_image.original_size_transform is not None
+                ):
+                    tgt_original_size_transform = (
+                        tgt_reg_image.original_size_transform
+                    )
+
                 if self.cache_images is True:
                     if reg_edge.get("override_prepro") is not None:
                         if (
@@ -626,16 +715,21 @@ class WsiReg2D(object):
                 )
 
                 reg_tforms = [sitk_pmap_to_dict(tf) for tf in reg_tforms]
+
                 if src_transforms is not None:
-                    initial_transforms = src_reg_image.transforms[0]
+                    initial_transforms = src_transforms[0]
                 else:
-                    initial_transforms = src_reg_image.transforms
+                    initial_transforms = src_reg_image.pre_reg_transforms
 
                 reg_edge["transforms"] = {
                     'initial': initial_transforms,
                     'registration': reg_tforms,
                 }
 
+                self.original_size_transforms.update(
+                    {tgt_name: tgt_original_size_transform}
+                )
+
                 reg_edge["registered"] = True
                 pmap_dict_to_json(
                     reg_edge["transforms"], str(output_path_tform)
@@ -650,16 +744,45 @@ class WsiReg2D(object):
 
     @transformations.setter
     def transformations(self, reg_graph_edges):
-        self._transformations = self._collate_transformations(reg_graph_edges)
+        self._transformations = self._collate_transformations()
 
-    def _collate_transformations(self, reg_graph_edges):
+    def add_merge_modalities(self, merge_name, modalities):
+        for modality in modalities:
+            try:
+                self.modalities[modality]
+            except KeyError:
+                raise ValueError(
+                    f"Modality for merger [{modality}] is not a modality "
+                    f"within the graph, current modalitles : "
+                    f"{self.modality_names}"
+                )
+            self.merge_modalities.update({merge_name: modalities})
+
+    def _generate_reg_transforms(self):
+        self._reg_graph_edges["reg_transforms"]
+
+    def _collate_transformations(self):
         transforms = {}
+        for reg_edge in self.reg_graph_edges:
+            if reg_edge["transforms"]["initial"] is not None:
+                initial_transforms = [
+                    RegTransform(t) for t in reg_edge["transforms"]["initial"]
+                ]
+            else:
+                initial_transforms = []
+            reg_edge["reg_transforms"] = {
+                'initial': initial_transforms,
+                'registration': [
+                    RegTransform(t)
+                    for t in reg_edge["transforms"]["registration"]
+                ],
+            }
         edge_modality_pairs = [v['modalities'] for v in self.reg_graph_edges]
         for modality, tform_edges in self.transform_paths.items():
             for idx, tform_edge in enumerate(tform_edges):
                 reg_edge_tforms = self.reg_graph_edges[
                     edge_modality_pairs.index(tform_edge)
-                ]["transforms"]
+                ]["reg_transforms"]
                 if idx == 0:
                     transforms[modality] = {
                         'initial': reg_edge_tforms['initial'],
@@ -670,7 +793,9 @@ class WsiReg2D(object):
 
         return transforms
 
-    def _transform_nonreg_image(self, modality_key, file_writer="ome.tiff"):
+    def _prepare_nonreg_image_transform(
+        self, modality_key, to_original_size=True
+    ):
         print(
             "transforming non-registered modality : {} ".format(modality_key)
         )
@@ -678,46 +803,52 @@ class WsiReg2D(object):
             self.project_name, modality_key
         )
         im_data = self.modalities[modality_key]
+        transformations = {"initial": None, "registered": None}
 
         if (
             im_data.get("preprocessing").get("rot_cc") is not None
             or im_data.get("preprocessing").get("flip") is not None
+            or im_data.get("preprocessing").get("mask_to_bbox") is True
+            or im_data.get("preprocessing").get("mask_bbox") is not None
+        ):
+
+            transformations.update(
+                {
+                    "initial": self._check_cache_modality(modality_key)[1][0],
+                    "registered": None,
+                }
+            )
+
+        if to_original_size is True:
+            transformations.update(
+                {"registered": self.original_size_transforms.get(modality_key)}
+            )
+
+        if (
+            transformations.get("initial") is None
+            and transformations.get("registered") is None
         ):
-            transformations = {
-                "initial": self._check_cache_modality(modality_key)[1][0],
-                "registered": None,
-            }
-        else:
             transformations = None
 
-        tfregimage = TransformRegImage(
-            output_path.stem,
-            im_data["image_filepath"],
-            im_data["image_res"],
-            transform_data=transformations,
-            channel_names=im_data.get("channel_names"),
-            channel_colors=im_data.get("channel_colors"),
-        )
-        im_fp = tfregimage.transform_image(
-            str(output_path.parent), output_type=file_writer, tile_size=512
-        )
-        return im_fp
+        return im_data, transformations, output_path
 
-    def _transform_image(
+    def _prepare_reg_image_transform(
         self,
         edge_key,
-        file_writer="ome.tiff",
         attachment=False,
         attachment_modality=None,
+        to_original_size=True,
     ):
         im_data = self.modalities[edge_key]
 
         if attachment is True:
             final_modality = self.reg_paths[attachment_modality][-1]
-            transformations = self.transformations[attachment_modality]
+            transformations = deepcopy(
+                self.transformations[attachment_modality]
+            )
         else:
             final_modality = self.reg_paths[edge_key][-1]
-            transformations = self.transformations[edge_key]
+            transformations = deepcopy(self.transformations[edge_key])
 
         print("transforming {} to {}".format(edge_key, final_modality))
 
@@ -726,21 +857,148 @@ class WsiReg2D(object):
             edge_key,
             final_modality,
         )
-        tfregimage = TransformRegImage(
-            output_path.stem,
+        if (
+            self.original_size_transforms.get(final_modality) is not None
+            and to_original_size is True
+        ):
+            original_size_transform = self.original_size_transforms[
+                final_modality
+            ]
+            transformations.update(
+                {"orig": [RegTransform(original_size_transform)]}
+            )
+
+        return im_data, transformations, output_path
+
+    def _transform_write_image(
+        self, im_data, transformations, output_path, file_writer="ome.tiff"
+    ):
+
+        tfregimage = reg_image_loader(
             im_data["image_filepath"],
             im_data["image_res"],
-            transform_data=transformations,
             channel_names=im_data.get("channel_names"),
             channel_colors=im_data.get("channel_colors"),
         )
-        im_fp = tfregimage.transform_image(
-            str(self.output_dir), output_type=file_writer, tile_size=512
-        )
+
+        ometiffwriter = OmeTiffWriter(tfregimage)
+
+        if transformations:
+            (
+                composite_transform,
+                itk_transforms,
+                final_transform,
+            ) = prepare_wsireg_transform_data(transformations)
+        else:
+            composite_transform, itk_transforms, final_transform = (
+                None,
+                None,
+                None,
+            )
+
+        if (
+            file_writer == "ome.tiff-bytile"
+            and ometiffwriter.reg_image.reader not in ["czi", "sitk"]
+        ):
+            im_fp = ometiffwriter.write_image_by_tile(
+                output_path.stem,
+                itk_transforms=itk_transforms,
+                composite_transform=composite_transform,
+                final_transform=final_transform,
+                output_dir=str(self.output_dir),
+            )
+        else:
+            im_fp = ometiffwriter.write_image_by_plane(
+                output_path.stem,
+                composite_transform=composite_transform,
+                final_transform=final_transform,
+                output_dir=str(self.output_dir),
+            )
 
         return im_fp
 
-    def transform_images(self, file_writer="ome.tiff", transform_non_reg=True):
+    def _transform_write_merge_images(self, to_original_size=True):
+        for merge_name, sub_images in self.merge_modalities.items():
+            im_fps = []
+            im_res = []
+            im_ch_names = []
+            transformations = []
+            final_modalities = []
+            for sub_image in sub_images:
+                im_data = self.modalities[sub_image]
+                im_fps.append(im_data["image_filepath"])
+                im_res.append(im_data["image_res"])
+                im_ch_names.append(im_data.get("channel_names"))
+
+                try:
+                    transforms = deepcopy(self.transformations[sub_image])
+                except KeyError:
+                    transforms = None
+
+                try:
+                    final_modalities.append(self.reg_paths[sub_image][-1])
+                except KeyError:
+                    initial_transforms = self._check_cache_modality(sub_image)[
+                        1
+                    ][0]
+                    if initial_transforms:
+                        initial_transforms = [
+                            RegTransform(t) for t in initial_transforms
+                        ]
+                        final_modalities.append(sub_image)
+                        transforms = {"initial": initial_transforms}
+                    else:
+                        transforms = None
+
+                transformations.append(transforms)
+
+            if all(final_modalities):
+                final_modality = final_modalities[0]
+            else:
+                raise ValueError("final modalities do not match on merge")
+
+            if (
+                self.original_size_transforms.get(final_modality) is not None
+                and to_original_size is True
+            ):
+                original_size_transform = self.original_size_transforms[
+                    final_modality
+                ]
+                for transformation in transformations:
+                    if transformation is None:
+                        transformation = {}
+                    transformation.update(
+                        {"orig": [RegTransform(original_size_transform)]}
+                    )
+
+            output_path = self.output_dir / "{}-{}_merged-registered".format(
+                self.project_name,
+                merge_name,
+            )
+
+            merge_regimage = MergeRegImage(
+                im_fps,
+                im_res,
+                channel_names=im_ch_names,
+            )
+
+            merge_ometiffwriter = MergeOmeTiffWriter(merge_regimage)
+
+            im_fp = merge_ometiffwriter.merge_write_image_by_plane(
+                output_path.stem,
+                sub_images,
+                transformations=transformations,
+                output_dir=str(self.output_dir),
+            )
+            return im_fp
+
+    def transform_images(
+        self,
+        file_writer="ome.tiff",
+        transform_non_reg=True,
+        remove_merged=True,
+        to_original_size=True,
+    ):
         """
         Transform and write images to disk after registration. Also transforms all attachment images
 
@@ -749,34 +1007,105 @@ class WsiReg2D(object):
         file_writer : str
             output type to use, sitk writes a single resolution tiff, "zarr" writes an ome-zarr multiscale
             zarr store
+        transform_non_reg : bool
+            whether to write the images that aren't transformed during registration as well
+        remove_merged: bool
+            whether to remove images that are stored in merged store, if True, images that are merged
+            will not be written as individual images as well
+        to_original_size: bool
+            write images that have been cropped for registration back to their original coordinate space
         """
-
         image_fps = []
 
         if all(
             [reg_edge.get("registered") for reg_edge in self.reg_graph_edges]
         ):
-            for key in self.reg_paths.keys():
-                self._transform_image(key, file_writer=file_writer)
+            # prepare workflow
+            merge_modalities = []
+            if len(self.merge_modalities.keys()) > 0:
+                for k, v in self.merge_modalities.items():
+                    merge_modalities.extend(v)
+
+            reg_path_keys = list(self.reg_paths.keys())
+            nonreg_keys = self._find_nonreg_modalities()
+
+            if remove_merged:
+                for merge_mod in merge_modalities:
+                    try:
+                        m_idx = reg_path_keys.index(merge_mod)
+                        reg_path_keys.pop(m_idx)
+                    except ValueError:
+                        pass
+                    try:
+                        m_idx = nonreg_keys.index(merge_mod)
+                        nonreg_keys.pop(m_idx)
+                    except ValueError:
+                        pass
+
+            for modality in reg_path_keys:
+                (
+                    im_data,
+                    transformations,
+                    output_path,
+                ) = self._prepare_reg_image_transform(
+                    modality,
+                    attachment=False,
+                    to_original_size=to_original_size,
+                )
+                im_fp = self._transform_write_image(
+                    im_data,
+                    transformations,
+                    output_path,
+                    file_writer=file_writer,
+                )
+                image_fps.append(im_fp)
 
             for (
                 modality,
                 attachment_modality,
             ) in self.attachment_images.items():
-
-                im_fp = self._transform_image(
+                (
+                    im_data,
+                    transformations,
+                    output_path,
+                ) = self._prepare_reg_image_transform(
                     modality,
-                    file_writer=file_writer,
                     attachment=True,
                     attachment_modality=attachment_modality,
+                    to_original_size=to_original_size,
+                )
+
+                im_fp = self._transform_write_image(
+                    im_data,
+                    transformations,
+                    output_path,
+                    file_writer=file_writer,
                 )
                 image_fps.append(im_fp)
+            if len(self.merge_modalities.items()) > 0:
+                im_fp = self._transform_write_merge_images(
+                    to_original_size=to_original_size
+                )
+                image_fps.append(im_fp)
+
         if transform_non_reg is True:
             # preprocess and save unregistered nodes
-            nonreg_keys = self._find_nonreg_modalities()
-
-            for key in nonreg_keys:
-                self._transform_nonreg_image(key, file_writer=file_writer)
+            for modality in nonreg_keys:
+                (
+                    im_data,
+                    transformations,
+                    output_path,
+                ) = self._prepare_nonreg_image_transform(
+                    modality,
+                    to_original_size=to_original_size,
+                )
+                im_fp = self._transform_write_image(
+                    im_data,
+                    transformations,
+                    output_path,
+                    file_writer=file_writer,
+                )
+                image_fps.append(im_fp)
 
         return image_fps
 
@@ -830,9 +1159,10 @@ class WsiReg2D(object):
                     final_modality,
                 )
             )
+            out_transforms = get_elastix_transforms(self.transformations[key])
 
             with open(output_path, 'w') as fp:
-                json.dump(self.transformations[key], fp, indent=4)
+                json.dump(out_transforms, fp, indent=4)
 
         for (
             modality,
@@ -850,8 +1180,12 @@ class WsiReg2D(object):
                 )
             )
 
+            out_transforms = get_elastix_transforms(
+                self.transformations[modality]
+            )
+
             with open(output_path, 'w') as fp:
-                json.dump(self.transformations[key], fp, indent=4)
+                json.dump(out_transforms, fp, indent=4)
 
     def add_data_from_config(self, config_filepath):
 
@@ -865,13 +1199,20 @@ class WsiReg2D(object):
                     if val.get("image_filepath") is not None
                     else val.get("image_filepath")
                 )
+
+                preprocessing = (
+                    "None"
+                    if val.get("preprocessing") is None
+                    else val.get("preprocessing")
+                )
+
                 self.add_modality(
                     key,
                     image_filepath,
                     image_res=val.get("image_res"),
                     channel_names=val.get("channel_names"),
                     channel_colors=val.get("channel_colors"),
-                    prepro_dict=val.get("preprocessing"),
+                    prepro_dict=preprocessing,
                     mask=val.get("mask"),
                 )
         else:
@@ -915,6 +1256,10 @@ class WsiReg2D(object):
             if all([re.get("registered") for re in self.reg_graph_edges]):
                 self.transformations = self.reg_graph_edges
 
+        if reg_config.get("merge_modalities") is not None:
+            for mn, mm in reg_config["merge_modalities"]:
+                self.add_merge_modalities(mn, mm)
+
     def reset_registered_modality(self, modalities):
         edge_keys = [
             r.get("modalities").get("source") for r in self.reg_graph_edges
@@ -927,7 +1272,7 @@ class WsiReg2D(object):
             self.reg_graph_edges[modality_idx]["registered"] = False
 
 
-if __name__ == "__main__":
+def main():
     import argparse
 
     def config_to_WsiReg2D(config_filepath):
@@ -974,3 +1319,9 @@ if __name__ == "__main__":
 
     if reg_graph.shape_sets:
         reg_graph.transform_shapes()
+
+
+if __name__ == "__main__":
+    import sys
+
+    sys.exit(main())
